#region History
// ----------------------------------------------------------------------------
//  History of file ConfusionMatrixView.cs.
//
//  Programmer          Date       Defect #       Description                  
//  =================   =========  ============   =============================
//  mgscott				15-Nov-05                 Created
//
//  Copyright (C) 2005 KLA-Tencor Corp, All Rights Reserved
//  KLA-Tencor Confidential--Special Handling
// ----------------------------------------------------------------------------
#endregion

using System;											// Used by EventArgs
using System.Collections;								// Used by ArrayList
using System.Windows.Forms;
using KLATencor.UIFramework;	
using KLATencor.RoadRunner.PF.UI.Infrastructure;		// needed for IContext
using KLATencor.RoadRunner.PF.BLIF.Types;				// needed for IDefectViewHandle
using KLATencor.RoadRunner.PF.BLIF.Results;
using KLATencor.RoadRunner.PF.BLIF.Results.DataSource;
using KLATencor.Infrastructure.Events;					// needed for IEventDistributor
using KLATencor.Infrastructure.TypedEvents;				// needed for IPublishingServices
using Syncfusion.ComponentModel;
using Syncfusion.Windows.Forms.Grid;
using KLATencor.RoadRunner.PF.BLIF.Infrastructure;

namespace  KLATencor.RoadRunner.PF.UI.Review.ConfusionMatrixView
{
	/// <summary>
	/// ConfusionMatrixView will create a grid based on two attributes.  Each cell in the
	/// grid will represent all defects that satisfy attribute value for both the row, and the column.
	/// </summary>
	[PluginInfoAttribute(Description = "Display defect grid representation of chart data", Platform = "")]
	[PluginPropertyAttribute(PluginPropertyClass="KLATencor.RoadRunner.PF.UI.Review.ConfusionMatrixView.ConfusionMatrixViewPropertyObj")]
	public class ConfusionMatrixView : System.Windows.Forms.UserControl, KLATencor.UIFramework.IView
	{
		public event KLATencor.UIFramework.ViewClosingEvent OnViewClosing;
		
		//Set this variable to your view's result. This information is returned in the ViewClosingEvent.
		private ViewClosingEventArgs.ViewResult viewResult = ViewClosingEventArgs.ViewResult.OK;

		private System.ComponentModel.IContainer components;	
		private IContext m_context  = null;
		// private IWaferResultsHandle m_waferResultsHandle = null;		
		private IDefectCollectionDS defectCollectionDS = null;
		private ISummarizeResultsDS m_summaryResultsDS=null;
		protected KTLogger m_Log	= new KTLogger("RR.UI.Review");

		private int binRows, gridRows;
		private int binCols, gridCols;
		private SummarizeBin[] rowBins;
		private SummarizeBin[] colBins;
		
		#region Designer Variables
		private System.Windows.Forms.Label labelTitle;
		private System.Windows.Forms.GroupBox groupBoxTitle;
		private Syncfusion.Windows.Forms.Tools.XPMenus.PopupMenusManager popupMenusManager1;
		private Syncfusion.Windows.Forms.Tools.XPMenus.PopupMenu popupMenu1;
		private Syncfusion.Windows.Forms.Tools.XPMenus.ParentBarItem parentBarItem1;
		private System.Windows.Forms.ImageList imageList1;
		private Syncfusion.Windows.Forms.Tools.XPMenus.BarItem barItemRowTotals;
		private Syncfusion.Windows.Forms.Tools.XPMenus.BarItem barItemColTotals;
		private Syncfusion.Windows.Forms.Tools.XPMenus.BarItem barItemShowAllDefects;
		private Syncfusion.Windows.Forms.Grid.GridControl gridCtl;
		#endregion

		/// <summary>
		/// ConfusionMatrixView Constructor
		/// </summary>
		public ConfusionMatrixView()
		{
			this.HasRowTotal = true;
			this.HasColTotal = true;
			this.RowAttr = "PWQRanking";
			this.ColAttr = "PWQBin";
			this.components = new System.ComponentModel.Container();
			InitializeComponent();					// Required for Windows Form Designer support
			ReceiveEvent();
//			OnInitView(this);
		}

		#region ConfusionMatrix Properties

		private string m_strRowAttributeName;
		public string RowAttr
		{
			get { return this.m_strRowAttributeName; }
			set { this.m_strRowAttributeName = value; }
		}

		private string m_strColAttributeName;
		public string ColAttr
		{
			get { return this.m_strColAttributeName; }
			set { this.m_strColAttributeName = (value==null ? string.Empty : value); }
		}

		private bool m_bHasRowTotal = false;
		public bool HasRowTotal
		{
			get { return this.m_bHasRowTotal; }
			set { this.m_bHasRowTotal = value; }
		}
		
		private bool m_bHasColTotal = false;
		public bool HasColTotal
		{
			get { return this.m_bHasColTotal; }
			set { this.m_bHasColTotal = value; }
		}
		#endregion

		protected override void Dispose( bool disposing )
		{
			if (disposing && components != null)
				components.Dispose();

			base.Dispose( disposing );
		}


		#region Component Designer generated code
		/// <summary> 
		/// Required method for Designer support - do not modify 
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(ConfusionMatrixView));
			Syncfusion.Windows.Forms.Grid.GridStyleInfo gridStyleInfo1 = new Syncfusion.Windows.Forms.Grid.GridStyleInfo();
			this.gridCtl = new Syncfusion.Windows.Forms.Grid.GridControl();
			this.popupMenusManager1 = new Syncfusion.Windows.Forms.Tools.XPMenus.PopupMenusManager(this.components);
			this.popupMenu1 = new Syncfusion.Windows.Forms.Tools.XPMenus.PopupMenu(this.components);
			this.parentBarItem1 = new Syncfusion.Windows.Forms.Tools.XPMenus.ParentBarItem();
			this.imageList1 = new System.Windows.Forms.ImageList(this.components);
			this.barItemRowTotals = new Syncfusion.Windows.Forms.Tools.XPMenus.BarItem();
			this.barItemColTotals = new Syncfusion.Windows.Forms.Tools.XPMenus.BarItem();
			this.barItemShowAllDefects = new Syncfusion.Windows.Forms.Tools.XPMenus.BarItem();
			this.groupBoxTitle = new System.Windows.Forms.GroupBox();
			this.labelTitle = new System.Windows.Forms.Label();
			((System.ComponentModel.ISupportInitialize)(this.gridCtl)).BeginInit();
			this.groupBoxTitle.SuspendLayout();
			this.SuspendLayout();
			// 
			// gridCtl
			// 
			this.gridCtl.AllowSelection = ((Syncfusion.Windows.Forms.Grid.GridSelectionFlags)((((Syncfusion.Windows.Forms.Grid.GridSelectionFlags.Row | Syncfusion.Windows.Forms.Grid.GridSelectionFlags.Column) 
				| Syncfusion.Windows.Forms.Grid.GridSelectionFlags.Table) 
				| Syncfusion.Windows.Forms.Grid.GridSelectionFlags.Cell)));
			this.gridCtl.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.gridCtl.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D;
			this.gridCtl.ControllerOptions = ((Syncfusion.Windows.Forms.Grid.GridControllerOptions)(((Syncfusion.Windows.Forms.Grid.GridControllerOptions.ClickCells | Syncfusion.Windows.Forms.Grid.GridControllerOptions.SelectCells) 
				| Syncfusion.Windows.Forms.Grid.GridControllerOptions.ExcelLikeSelection)));
			this.gridCtl.Data = ((Syncfusion.Windows.Forms.Grid.GridData)(resources.GetObject("gridCtl.Data")));
			this.gridCtl.DragSelectedCellsMouseButtonsMask = System.Windows.Forms.MouseButtons.None;
			this.gridCtl.EnterKeyBehavior = Syncfusion.Windows.Forms.Grid.GridDirectionType.None;
			this.gridCtl.Location = new System.Drawing.Point(0, 50);
			this.gridCtl.Name = "gridCtl";
			this.gridCtl.ResizeColsBehavior = Syncfusion.Windows.Forms.Grid.GridResizeCellsBehavior.None;
			this.gridCtl.ResizeRowsBehavior = Syncfusion.Windows.Forms.Grid.GridResizeCellsBehavior.None;
			this.gridCtl.RowCount = 1000;
			this.gridCtl.SelectCellsMouseButtonsMask = System.Windows.Forms.MouseButtons.Left;
			this.gridCtl.Size = new System.Drawing.Size(720, 430);
			this.gridCtl.TabIndex = 0;
			gridStyleInfo1.CellAppearance = Syncfusion.Windows.Forms.Grid.GridCellAppearance.Flat;
			gridStyleInfo1.CellType = "Static";
			gridStyleInfo1.HorizontalAlignment = Syncfusion.Windows.Forms.Grid.GridHorizontalAlignment.Center;
			gridStyleInfo1.ReadOnly = true;
			gridStyleInfo1.ShowButtons = Syncfusion.Windows.Forms.Grid.GridShowButtons.Show;
			this.gridCtl.TableStyle = gridStyleInfo1;
			this.gridCtl.Text = "gridCtl";
			this.popupMenusManager1.SetXPContextMenu(this.gridCtl, this.popupMenu1);
			this.gridCtl.PrepareViewStyleInfo += new Syncfusion.Windows.Forms.Grid.GridPrepareViewStyleInfoEventHandler(this.PrepareViewStyleInfoEventHandler);
			this.gridCtl.SelectionChanging += new Syncfusion.Windows.Forms.Grid.GridSelectionChangingEventHandler(this.SelectionChangingEventHandler);
			this.gridCtl.CellClick += new Syncfusion.Windows.Forms.Grid.GridCellClickEventHandler(this.CellClickEventHandler);
			this.gridCtl.DrawCellDisplayText += new Syncfusion.Windows.Forms.Grid.GridDrawCellDisplayTextEventHandler(this.DrawCellDisplayTextEventHandler);
			// 
			// popupMenu1
			// 
			this.popupMenu1.ParentBarItem = this.parentBarItem1;
			// 
			// parentBarItem1
			// 
			this.parentBarItem1.CategoryIndex = -1;
			this.parentBarItem1.ID = "";
			this.parentBarItem1.ImageIndex = 0;
			this.parentBarItem1.ImageList = this.imageList1;
			this.parentBarItem1.Items.AddRange(new Syncfusion.Windows.Forms.Tools.XPMenus.BarItem[] {
																										this.barItemRowTotals,
																										this.barItemColTotals,
																										this.barItemShowAllDefects});
			this.parentBarItem1.SeparatorIndices.AddRange(new int[] {
																		2});
			// 
			// imageList1
			// 
			this.imageList1.ImageSize = new System.Drawing.Size(16, 16);
			this.imageList1.TransparentColor = System.Drawing.Color.Transparent;
			// 
			// barItemRowTotals
			// 
			this.barItemRowTotals.CategoryIndex = -1;
			this.barItemRowTotals.ID = "RowTotals";
			this.barItemRowTotals.Text = "Row Totals";
			this.barItemRowTotals.Click += new System.EventHandler(this.OnClick_MenuRowTotals);
			// 
			// barItemColTotals
			// 
			this.barItemColTotals.CategoryIndex = -1;
			this.barItemColTotals.ID = "ColTotals";
			this.barItemColTotals.Text = "Column Totals";
			this.barItemColTotals.Click += new System.EventHandler(this.OnClick_MenuColTotals);
			// 
			// groupBoxTitle
			// 
			this.groupBoxTitle.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.groupBoxTitle.Controls.Add(this.labelTitle);
			this.groupBoxTitle.Location = new System.Drawing.Point(24, 0);
			this.groupBoxTitle.Name = "groupBoxTitle";
			this.groupBoxTitle.Size = new System.Drawing.Size(672, 40);
			this.groupBoxTitle.TabIndex = 1;
			this.groupBoxTitle.TabStop = false;
			// 
			// labelTitle
			// 
			this.labelTitle.Anchor = ((System.Windows.Forms.AnchorStyles)(((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.labelTitle.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.labelTitle.Location = new System.Drawing.Point(16, 16);
			this.labelTitle.Name = "labelTitle";
			this.labelTitle.Size = new System.Drawing.Size(640, 16);
			this.labelTitle.TabIndex = 0;
			this.labelTitle.Text = "label1";
			this.labelTitle.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// ConfusionMatrixView
			// 
			this.Controls.Add(this.groupBoxTitle);
			this.Controls.Add(this.gridCtl);
			this.Name = "ConfusionMatrixView";
			this.Size = new System.Drawing.Size(720, 480);
			((System.ComponentModel.ISupportInitialize)(this.gridCtl)).EndInit();
			this.groupBoxTitle.ResumeLayout(false);
			this.ResumeLayout(false);

		}
		#endregion

		#region Implementation of IView
		public bool CanCloseView() 
		{
			// TODO: if you need to do something before your view closes do it here (e.g. save data).
			return true; 
		} 
		
		public void CloseView() 
		{
			if (OnViewClosing!=null) 
				OnViewClosing(this, new ViewClosingEventArgs(viewResult));
			
			if(null != this.m_context)
			{
				this.m_context.Changed -= new EventHandler(ContextChanged);
			}

			if (m_summaryResultsDS!=null)
			{
				m_summaryResultsDS.Dispose();
				m_summaryResultsDS=null; 
			}

			UnsubcribeEvent();
			UnsubscribeEvents();
			DeleteAndUnitializeDS();
		}

		//public void SetPropertyObj ( object propertyObj ) { }
		private ConfusionMatrixViewPropertyObj propertyObj = new ConfusionMatrixViewPropertyObj ( );
		public void SetPropertyObj ( object propertyObj )
		{
			if (propertyObj != null)
			{
				this.propertyObj = (ConfusionMatrixViewPropertyObj)propertyObj;
			}
		}

		
		public void OnInitView ( object viewDoc )
		{
			this.m_context  = Context.ActiveContext;
			this.m_context.Changed += new EventHandler(ContextChanged);			 
			ContextChanged( this, new EventArgs() );
			this.barItemRowTotals.Checked = this.HasRowTotal;
			this.barItemColTotals.Checked = this.HasColTotal;
			SubscribeEvents();
		}
		
		public void OnShowView(bool show) {}
		
		public object[] GetToolbar(ElementType type) 
		{ 
			return (null); 
		}
	
		public object GetMenu(ElementType type) { return (null); }
		
		// TODO: add your static handlers to the handler map new HandlerMapEntry (typeof(YourId), new YourHandler())
		private HandlerMap handlerMap = new HandlerMap (new HandlerMapEntry[] {}); 
		public object GetHandlerMap() 
		{ 
			return (handlerMap); 
		}
		#endregion

		#region IView Event Handlers
		private void ReleaseDefectView()
		{
			if(this.defectView != null)
			{
				this.defectView.Release();
				this.defectView = null;
				this.Visible = false;
			}
		}

		private void UnsubscribeEvents()
		{
			this.defectCollectionDS.DefectChanged -= new DefectChangedEventHandler(DefectChangedEventHandler);
			this.defectCollectionDS.DatasourceChanged -= new DatasourceChangedEventHandler(defectCollectionDS_DatasourceChanged);
		}

		private void SubscribeEvents()
		{
			this.defectCollectionDS.DefectChanged += new DefectChangedEventHandler(DefectChangedEventHandler);
			this.defectCollectionDS.DatasourceChanged += new DatasourceChangedEventHandler(defectCollectionDS_DatasourceChanged);
		}
		
		IDefectViewHandle defectView = null;
		
		private void DeleteAndUnitializeDS()
		{
			if(null != this.defectCollectionDS)
			{
				this.defectCollectionDS.DefectChanged -= new DefectChangedEventHandler(DefectChangedEventHandler);
				IDisposable disposable = this.defectCollectionDS as IDisposable;
				if(null != disposable)
				{
					disposable.Dispose();
				}
				this.defectCollectionDS = null;
			}
			ReleaseDefectView();
		}

		private void ContextChanged(object sender, EventArgs args)
		{ 
			DeleteAndUnitializeDS();

			if (this.defectView != null)
				this.defectView.Release();

			this.defectView = Context.ActiveContext.GetContextHandle (this.propertyObj.ContextKey) as IDefectViewHandle;
			if(this.defectView != null)  //defect data is finally available
			{ 
				// this.m_waferResultsHandle = defectView.GetInnerHandle(typeof(IWaferResultsHandle)) as IWaferResultsHandle;
				this.defectView.Acquire();
				this.Visible = true;
				this.defectCollectionDS = (IDefectCollectionDS)BL.DefaultInstance.DatasourceFactory.GetDatasource(typeof(IDefectCollectionDS));
				this.defectCollectionDS.Initialize(defectView);
//				UnsubscribeEvents();
//				SubscribeEvents();
				UpdateCM();
			}
			else
			{
				this.Visible = false;
			}
		}                
		#endregion

		private void UpdateCM()
		{
			this.m_Log.LogTrace("ConfusionMatrixView::UpdateCM:Begin");
			try
			{
				// Update Title Box
				UpdateTitle();

				int numDefects = this.defectCollectionDS.Count;
				if (numDefects == 0)
				{
					this.gridCtl.BeginUpdate();
					this.gridCtl.RowCount = 1;
					this.gridCtl.ColCount = 1; 
					int [,] grid = new int[2, 2];
					for (int i = 0; i < 2; i++)
						for (int j = 0; j < 2; j++)
                            grid[i,j] = 0;
					this.gridCtl.Model.PopulateValues( GridRangeInfo.Cells(0, 0, 2, 2), grid );	
					this.gridCtl[0,1].Text = "Totals";			// Column Total
					this.gridCtl[1,0].Text = "Totals";			// Row Total
					this.gridCtl.Model.ColWidths.ResizeToFit( GridRangeInfo.Cells(0, 0, 2, 2));
					this.gridCtl.EndUpdate();
					this.gridCtl.Refresh();
					this.gridCtl.Invalidate();
					this.Refresh();
					return;
				}
			
				//Get the column values
				SummarizeResultsCmd localResultsCmd = new SummarizeResultsCmd(
					GetWaferResultsHandle(this.defectView), ColAttr, null, null, null,false,false);
				ISummarizeResultsDS localResultsDS = (ISummarizeResultsDS)BL.DefaultInstance.Execute(localResultsCmd);
				colBins = localResultsDS.CountBins; 
				this.binCols = colBins.Length;
			
				
				// Get the row values
				localResultsCmd = new SummarizeResultsCmd(
					GetWaferResultsHandle(this.defectView), RowAttr, null, null, null,false,false);
				localResultsDS = (ISummarizeResultsDS)BL.DefaultInstance.Execute(localResultsCmd);
				rowBins = localResultsDS.CountBins; 
				this.binRows = rowBins.Length;

				// Add 1 for Row / Column Total, and another for the Row / Column Headers
				this.gridRows = binRows + 2;
				this.gridCols = binCols + 2;

				// Begin Updating Confusion Matrix
				this.gridCtl.BeginUpdate();

				// Reset data (removes last row/column totals)
				this.gridCtl.RowCount = 1;
				this.gridCtl.ColCount = 1;

				// Reset size of grid
				this.gridCtl.RowCount = this.gridRows-1;
				this.gridCtl.ColCount = this.gridCols-1;				
				int [,] gridArray = new int[this.gridRows, this.gridCols];

				// Set Row borders - don't forget that row zero is the row header
				for (int i = 0; i < this.binRows; i++)
					gridArray[i+1,0] = Convert.ToInt32(rowBins[i].Max);

				// Set Column Borders - don't forget that col zero is the column header
				for (int j = 0; j < this.binCols; j++)
					gridArray[0,j+1] = Convert.ToInt32(colBins[j].Max);

				//
				// Draw Matrix
				//

				// int numDefects = this.defectCollectionDS.Count;

				// Populate the Confusion Matrix
				for (int i = 0; i < numDefects; i++)
				{
					// 1.  Grab each defect
					if (this.defectCollectionDS.DefectAccessor.MoveToByIndex(i))
					{
						// 2.  Get The X / Y attribute value for this defect
						int x = Convert.ToInt32(this.defectCollectionDS.DefectAccessor[RowAttr]);
						int y = Convert.ToInt32(this.defectCollectionDS.DefectAccessor[ColAttr]);

						// 3.  get the index into the matrix for these values
						int idxX = FindHeaderIdx(rowBins, x);
						int idxY = FindHeaderIdx(colBins, y);

						// 4.  Increment current cell - don't forget that row/col zero is the header row/col
						gridArray[idxX,idxY]++;
					}
				}

				//
				// Formatting of Row / Column Total rows handled by PrepareViewStyleInfoEventHandler
				//

				// 5.  Add Column Totals
				for (int i = 0; i < this.binRows; i++)
				{
					int nColTotal = 0;
					for (int j = 0; j < this.binCols; j++)
						nColTotal += gridArray[i+1,j+1];
					gridArray[i+1,gridCols-1] = nColTotal;
				}

				// 6.  Add Row Totals
				for (int i = 0; i < this.binCols; i++)
				{
					int nRowTotal = 0;
					for (int j = 0; j < this.binRows; j++)
						nRowTotal += gridArray[j+1,i+1];
					gridArray[gridRows-1,i+1] = nRowTotal;
				}

				// 8.  Fill out bottom/right cell
				gridArray[gridRows-1,gridCols-1] = numDefects;

				// 9.  Populate matrix
				this.gridCtl.Model.PopulateValues( GridRangeInfo.Cells(0, 0, this.gridRows, this.gridCols), gridArray );	

				// 10.  Add text to grid
				this.gridCtl[0,gridCols-1].Text = "Totals";	// Column Total
				this.gridCtl[gridRows-1,0].Text = "Totals";	// Row Total

				// 11.  Set column width
				this.gridCtl.Model.ColWidths.ResizeToFit( GridRangeInfo.Cells(0, 0,this.gridRows+1, this.gridCols+1));

				// 12.  Hide row / column totals if need be
				this.gridCtl.RowHeights[gridRows-1] = HasColTotal ? this.gridCtl.RowHeights[0] : 0;
				this.gridCtl.ColWidths[gridCols-1] = HasRowTotal ? this.gridCtl.ColWidths[0] : 0;

				this.gridCtl.EndUpdate();
				this.gridCtl.Refresh();

				localResultsDS.Dispose();
				localResultsDS = null;
			}
			catch
			{
//				System.Diagnostics.Debug.Assert(false,"Exception thrown in ConfusionMatrixView.UpdateCM");
			}

			this.m_Log.LogTrace("ConfusionMatrixView::UpdateCM:End");
		}


		/// <summary>
		/// Update the Title box in the Confusion Matrix.
		/// </summary>
		private bool UpdateTitle()
		{
			string strRowTitle = "", strColTitle = "";
			bool bRowTitleFound = false, bColTitleFound = false;
			DefectFieldInfoCollection defectFieldInfoList =  this.defectCollectionDS.DefectFieldInfoCollection ;
			for(int i = 0; i < defectFieldInfoList.Count; ++i)
			{
				if (bRowTitleFound && bColTitleFound)
					break;

				string strFieldName = defectFieldInfoList[i].Name;
				string strDisplayName = defectFieldInfoList[i].DisplayName;

				if (!bRowTitleFound && (this.RowAttr == strFieldName))
				{
					strRowTitle = strDisplayName;
					bRowTitleFound = true;
				}
				if (!bColTitleFound && (this.ColAttr == strFieldName))
				{
					strColTitle = strDisplayName;
					bColTitleFound = true;
				}
			}

			if ( (bRowTitleFound == true) && (bColTitleFound == true) )
			{
				string strTitle = strRowTitle + "    X    " + strColTitle;
				this.labelTitle.Text = strTitle;
			}
			else
			{
				this.labelTitle.Text = "Row and/or column attribute missing or invalid";
				// MessageBox.Show("Row or Column Attribute not found in list of Attributes.", "PWQ Binning View Matrix");
			}
			return ( (bRowTitleFound == true) && (bColTitleFound==true) );
		}

		/// <summary>
		/// If a cell/row/column is selected, pass this iformation on to the Read Only (Attribute) FolderListCtrl
		/// The FolderListCtrl will change its combo box to FolderListCtlAttr, select the folder associated with AttrValue,
		/// then tell the ImageGallery to show all images for that particular value, grouped by GalleryGroupByAttr
		/// </summary>
		/// <param name="AttrStr">String representing the attribute the ReadOnly FolderListCtrl needs to be set to</param>
		/// <param name="AttrValue">Number representing the folder that needs to be selected (and reflected in the Image Gallery</param>
		/// <param name="GroupByAttr">String representing the GroupBy attribute in the Image Gallery</param>
		/// <param name="GroupByValue">Number representing the group we need to scroll to if bScrollGallery is set to true</param>
		/// <param name="bScrollGallery">flag determining whether we need to scroll to that particular AttrValue in the Image Gallery</param>
#if MultiSelect
		private void UpdateAttributeFolderListCtrl(string AttrStr, ArrayList AttrValueList, string GroupByAttr, ArrayList GroupByValueList, bool bScrollGallery)
		{
			try
			{
				IEventDistributor distributor = EventDistributorFactory.GetEventDistributor(EventModel.Typed);
				IPublishingServices publisher = distributor as IPublishingServices;	
				if (publisher != null)
				{
					string strAttrDisplayStr = GetAttributeDisplayName(AttrStr);
					string strGroupByAttrDisplayStr = GetAttributeDisplayName(GroupByAttr);

					// Post the Event to the Event Distributor
					publisher.PostEvent(
						new KLATencor.RoadRunner.PF.UI.Types.MatrixSelectionChangedEvent(AttrStr, strAttrDisplayStr, AttrValueList, 
								GroupByAttr, strGroupByAttrDisplayStr, GroupByValueList, bScrollGallery), "MatrixSelectionChanged");
				}
			}
			catch
			{
				// Do nothing
				// System.Diagnostics.Debug.Assert(false,"Exception thrown in ConfusionMatrixView.UpdateAttributeFolderListCtrl");
			}
		}
#else
		private void UpdateAttributeFolderListCtrl(string AttrStr, int AttrValue, string GroupByAttr, int GroupByValue, bool bScrollGallery)
		{
			try
			{
				IEventDistributor distributor = EventDistributorFactory.GetEventDistributor(EventModel.Typed);
				IPublishingServices publisher = distributor as IPublishingServices;	
				if (publisher != null)
				{
					string strAttrDisplayStr = GetAttributeDisplayName(AttrStr);
					string strGroupByAttrDisplayStr = GetAttributeDisplayName(GroupByAttr);
					
					// Post the Event to the Event Distributor
					publisher.PostEvent(
						new KLATencor.RoadRunner.PF.UI.Types.MatrixSelectionChangedEvent(AttrStr, strAttrDisplayStr, AttrValue, 
								GroupByAttr, strGroupByAttrDisplayStr, GroupByValue, bScrollGallery), "MatrixSelectionChanged");
				}
			}
			catch
			{
				// Do nothing
				// System.Diagnostics.Debug.Assert(false,"Exception thrown in ConfusionMatrixView.UpdateAttributeFolderListCtrl");
			}
		}
#endif

		private string GetAttributeDisplayName(string strAttrName)
		{
			DefectFieldInfoCollection defectFieldInfoList =  this.defectCollectionDS.DefectFieldInfoCollection ;
			for(int i = 0; i < defectFieldInfoList.Count; ++i)
			{
				string strFieldName = defectFieldInfoList[i].Name;
				string strDisplayName = defectFieldInfoList[i].DisplayName;
				if (strAttrName == strFieldName)
				{
					return strDisplayName;
				}
			}
			return strAttrName;
		}

		#region Event Handlers

		/// <summary>
		/// This event forces only a single cell, row, or column to be selected.
		/// No muti-cell, multi-row, or multi-column selections allowed
		/// </summary>
		/// <param name="sender">Sent by the GridSelectionChangingEventArgs event</param>
		/// <param name="e">Sent by the GridSelectionChangingEventArgs event</param>
		private void SelectionChangingEventHandler(object sender, Syncfusion.Windows.Forms.Grid.GridSelectionChangingEventArgs e)
		{
			try
			{
				GridRangeInfo range = e.Range;
				GridRangeInfoType rType = e.Range.RangeType;

				// If selection is the top/left cell, don't allow it.
				// Still need to disable the "selection" of that cell though, 
				// which is done in PrepareViewStyleInfoEventHandler
				if (rType == GridRangeInfoType.Table)
					e.Cancel = true;
#if MultiSelect
#else
				// if selection is multirow, don't allow it
				else if ( (rType == GridRangeInfoType.Rows) && (range.Height > 1) )
				{
					e.Cancel = true;
					MessageBox.Show("only a single row is allowed.");
				}
#endif
					// If there are Row totals, and selection is Row Total, don't allow it - no error required
				else if ( range.Bottom == this.gridRows - 1 )
					e.Cancel = true;

					// if there are Col totals, and selection is a col total, don't allow it - no error required
				else if ( range.Right == this.gridCols - 1 )
					e.Cancel = true;

#if MultiSelect
#else
					// If selection is multicolumn, don't allow it
				else if ( (rType == GridRangeInfoType.Cols) && (range.Width > 1 ) )
				{
					e.Cancel = true;
					MessageBox.Show("Only a single column is allowed.");
				}
#endif
					
					// If selection is multicell, don't allow it
				else if ( (rType == GridRangeInfoType.Cells) && 
					( (range.Height > 1) || (range.Width > 1) ) )
				{
					e.Cancel = true;
					MessageBox.Show("Only a single cell is allowed.");
				}

#if MultiSelect
#else
				// If selection is non-contiguous, clear selection
				if(this.gridCtl.Selections.Count > 1)
					this.gridCtl.Selections.Clear();				
#endif
			}
			catch
			{
				System.Diagnostics.Debug.Assert(false,"Exception thrown in ConfusionMatrixView.SelectionChangingEventHandler");
			}
		}

		/// <summary>
		/// This event sends the defects represented by the current cell/row/column to be sent to the Image Gallery.
		/// Because of the OnSelectionChanging event above, we should only have a single cell, row, or column
		/// </summary>
		/// <param name="sender">Sent by the GridCellClickEventArgs event</param>
		/// <param name="e">Sent by the GridCellClickEventArgs event</param>
		private void CellClickEventHandler(object sender, Syncfusion.Windows.Forms.Grid.GridCellClickEventArgs e)
		{
			// If there are row/col totals, and we are over one, do nothing
			if ( e.RowIndex == this.gridRows - 1)
				return;

			if ( e.ColIndex == this.gridCols - 1)
				return;

			this.Cursor = Cursors.WaitCursor;
			try
			{
				GridRangeInfo rangeInfo = new GridRangeInfo();
				rangeInfo.ExpandRange(0, 0, this.gridCtl.RowCount, this.gridCtl.ColCount);
				GridRangeInfoList rangeList;

				int nRowVal = -1, nColVal = -1;
				bool bConsiderCurrentCell = true;
				if(this.gridCtl.Selections.GetSelectedRanges (out rangeList, bConsiderCurrentCell))
				{ 
#if MultiSelect
					ArrayList rowList = null;
					rowList = new ArrayList();
					ArrayList colList = null;
					colList = new ArrayList();
					bool bFromColumns = true;
					bool bScrollGallery = false;
#else
					// Clear selection in grid, etc... first
					if( this.defectView != null )
					{
						SelectDefectsByIdCmd selectDefectsCmd = new SelectDefectsByIdCmd(
							this.defectView.GetInnerHandle(typeof(IWaferResultsHandle)) as IWaferResultsHandle, null, false );
						BL.DefaultInstance.Execute( selectDefectsCmd );
					}
#endif
					// Should only be 1 range.
					foreach(GridRangeInfo currentRange in rangeList) 
					{
						switch(currentRange.RangeType) 
						{
								// Clear filtered list, then filter on the current cell
							case GridRangeInfoType.Cells: 
								// If the current cell is blank, fall through
								string curCell = this.gridCtl[currentRange.Top,currentRange.Left].CellValue.ToString();
								if (curCell != "0")
								{
									nRowVal = FindHeaderFromIdx( rowBins, currentRange.Top-1 );
									nColVal = FindHeaderFromIdx( colBins, currentRange.Left-1 );

									// If top/left cell, fall through - we now have all defects
									if ( (nRowVal >= 0) && (nColVal >= 0) )
									{
#if MultiSelect
										bScrollGallery = true;
										colList.Add(nColVal);
#else
										UpdateAttributeFolderListCtrl(ColAttr, nColVal, RowAttr, nRowVal, true);
#endif
									}
								}
								break;

								// Clear filtered list, then filter on the current column
							case GridRangeInfoType.Cols: 
								nColVal = FindHeaderFromIdx( colBins, currentRange.Left-1 );
								if (nColVal >= 0)
								{
#if MultiSelect
									colList.Add(nColVal);
#else
									UpdateAttributeFolderListCtrl(ColAttr, nColVal, RowAttr, nRowVal, false);
#endif
								}
								break; 

								// Clear filtered list, then filter on the current row
							case GridRangeInfoType.Rows:
								nRowVal = FindHeaderFromIdx( rowBins, currentRange.Top-1 );
								if (nRowVal >= 0)
								{
#if MultiSelect
									bFromColumns = false;
									rowList.Add(nRowVal);
#else
									UpdateAttributeFolderListCtrl(RowAttr, nRowVal, ColAttr, nColVal, false);
#endif
								}
								break;
 
								case GridRangeInfoType.Table:
							default: 
								System.Diagnostics.Debug.Assert(false, "Invalid Range type selected in ConfusionMatrixView.CellClickEventHandler");
								break;
						}	// switch(currentRange.RangeType) 
					}	// foreach(GridRangeInfo currentRange in rangeList) 
#if MultiSelect
					UpdateAttributeFolderListCtrl(
						bFromColumns ? ColAttr : RowAttr, 
						bFromColumns ? colList : rowList,
						bFromColumns ? RowAttr : ColAttr, 
						bFromColumns ? rowList : colList,
						bScrollGallery);
#endif
				}	// if(this.gridCtl.Selections.GetSelectedRanges (out rangeList, bConsiderCurrentCell))
			}
			catch
			{
				System.Diagnostics.Debug.Assert(false,"Exception thrown in ConfusionMatrixView.CellClickEventHandler");
			}
			this.Cursor = Cursors.Arrow;
		}

		/// <summary>
		/// This Event handler will force the row / column totals to look different.
		/// It also causes the top/left cell to not depress when you click on it.
		/// NTR: Would like to turn Bold back on, but the column totals won't fit over 3 chars if you do.
		/// </summary>
		/// <param name="sender">Sent by the GridPrepareViewStyleInfoEvent event</param>
		/// <param name="e">Sent by the GridPrepareViewStyleInfoEvent event</param>
		private void PrepareViewStyleInfoEventHandler(object sender, Syncfusion.Windows.Forms.Grid.GridPrepareViewStyleInfoEventArgs e)
		{
			// Column Totals
			if ( e.RowIndex == this.gridCtl.RowCount )
			{
				//e.Style.Font.Bold = true;
				e.Style.Enabled = false;			
			}

			// Row Totals
			else if ( e.ColIndex == this.gridCtl.ColCount )
			{
				//e.Style.Font.Bold = true;
				e.Style.Enabled = false;			
			}		

			// Top/left cell
			else if (e.RowIndex == 0 && e.ColIndex == 0)
				e.Style.Enabled = false;
		}

		/// <summary>
		/// This Event Handler will catch any time a "DefectChangedEvent" event occurs
		/// </summary>
		/// <param name="sender">Sent by the DefectChangedEvent event</param>
		/// <param name="args">Sent by the DefectChangedEvent event</param>
		private void DefectChangedEventHandler(object sender, DefectChangedEventArgs args)
		{
			if(this.InvokeRequired)
			{
				this.BeginInvoke(new DefectChangedEventHandler(DefectChangedEventHandler),
					new object[]{ sender, args});
				return;
			}

			if (args != null)
			{
				if(args.DefectFieldNames.Count > 0)
				{
					string strField = args.DefectFieldNames[0].ToString();
				}
			}

			//UpdateCM();
		}

		private void defectCollectionDS_DatasourceChanged(object sender, DatasourceChangedEventArgs args)
		{
			if(args.Reason == DatasourceChangedReason.Updated)
			{
				if(this.InvokeRequired)
				{
					this.BeginInvoke((new DatasourceChangedEventHandler(defectCollectionDS_DatasourceChanged)),
						new object[]{sender, args});	
					return;
				}
				UpdateCM();
			}

		}

		/// <summary>
		/// This Event Handler will change any cell containing zero ("0") to blank.
		/// However, we DO want the row / column headers to contain zero, if there is one.
		/// </summary>
		/// <param name="sender">Sent by the DrawCellDisplayText event</param>
		/// <param name="args">Sent by the DrawCellDisplayText event</param>
		private void DrawCellDisplayTextEventHandler(object sender, GridDrawCellDisplayTextEventArgs e)
		{
			int col = e.Style.CellIdentity.ColIndex;
			int row = e.Style.CellIdentity.RowIndex;
			if ( ((col != 0) && (row != 0))	||	// if not a row / column header, OR
				((col == 0) && (row == 0)) )	// if not the top/left cell
			{
				if(e.DisplayText == "0")
				{
					e.DisplayText = "";
					e.Cancel = true;
				}
			}
		}

		private Guid subscriptionId;
		private bool eventSubscribed = false;
		/// <summary>
		/// Subscribing to Attribute / BreakdownAttribute changes in ChartView
		/// ConfusionMatrix needs to reflect these settings (X / Y Axes
		/// </summary>
		private void ReceiveEvent()
		{
			if(!this.eventSubscribed)
			{
				try
				{
					IEventDistributor distributor = EventDistributorFactory.GetEventDistributor(EventModel.Typed);
					ISubscriptionServices subscriber = distributor as ISubscriptionServices;
					subscriptionId = subscriber.RegisterSubscription();

					// PWQBinning (either Feature or Image-based) has just completed
					subscriber.SubscribeEvents(subscriptionId, "PWQBinning", new SubscriberEventHandler(PWQBinningCompleted));
					this.eventSubscribed = true;
				}
				catch(Exception ex)
				{
					this.m_Log.LogTrace( "Exception in EDS Subscription" + ex.Message );
				}
			}
		}

		private void UnsubcribeEvent()
		{
			if(this.eventSubscribed)
			{
				IEventDistributor distributor = EventDistributorFactory.GetEventDistributor(EventModel.Typed);
				ISubscriptionServices subscriber = distributor as ISubscriptionServices;
				subscriber.UnsubscribeEvents(subscriptionId, new SubscriberEventHandler(PWQBinningCompleted));
				subscriber.UnregisterSubscription(subscriptionId);
				this.eventSubscribed = false;
			}
		}
		
		/// <summary>
		/// Image-Based PWQBinning has just completed (either Feature or Image-based)
		/// Now we need to update the PWQBin Attribute list, if it's visible.
		/// </summary>
		/// <param name="info"></param>
		/// <param name="obj"></param>
		private delegate void PWQBinningCompletedDelegate(ReceivedEventInfo info, object obj);
		private void PWQBinningCompleted(ReceivedEventInfo info, object obj)
		{
			if (this.InvokeRequired)
			{
				this.BeginInvoke(new PWQBinningCompletedDelegate(PWQBinningCompleted), new object[] {info, obj});
				return;
			}
			UpdateCM();
		}

        /// <summary>
		/// Event handler for clicking on the menu pick "Column Totals".
		/// Change the property, show/hide the total, then update the menu pick
		/// </summary>
		/// <param name="sender">Sent by the Clicked event</param>
		/// <param name="e">Sent by the Clickedt event</param>
		private void OnClick_MenuColTotals(object sender, System.EventArgs e)
		{
			this.HasColTotal = !this.HasColTotal;
			this.gridCtl.RowHeights[gridRows-1] = HasColTotal ? this.gridCtl.RowHeights[0] : 0;		
			this.barItemColTotals.Checked = this.HasColTotal;
		}

		/// <summary>
		/// Event handler for clicking on the menu pick "Row Totals"
		/// Change the property, show/hide the total, then update the me
		/// </summary>
		/// <param name="sender">Sent by the Clicked event</param>
		/// <param name="e">Sent by the Clickedt event</param>
		private void OnClick_MenuRowTotals(object sender, System.EventArgs e)
		{
			this.HasRowTotal = !this.HasRowTotal;
			this.gridCtl.ColWidths[gridCols-1] = HasRowTotal ? this.gridCtl.ColWidths[0] : 0;
			this.barItemRowTotals.Checked = this.HasRowTotal;
		}

		#endregion

		#region Helper Functions

		/// <summary>
		/// Returns the Wafer Handle of the current defect view
		/// </summary>
		/// <param name="handle">Defect Handle</param>
		/// <returns>Wafer Handle of the current defect handle.</returns>
		private IWaferResultsHandle GetWaferResultsHandle(IDefectViewHandle handle)
		{
			IWaferResultsHandle waferHandle = null;
			if(null != handle)
				waferHandle = handle.GetInnerHandle(typeof(IWaferResultsHandle)) as IWaferResultsHandle;
			return waferHandle;
		}

		/// <summary>
		/// Returns the index of the current row / column (class code, etc...
		/// </summary>
		/// <param name="binHeader">Row or Column Header array</param>
		/// <param name="val">Row / Column title of the currently selected Row / Column</param>
		/// <returns>Index into the Row / Column array for that particular Title</returns>
		int FindHeaderIdx( SummarizeBin[] binHeader, int val )
		{
			try
			{
				for(int i = 0; i < binHeader.Length; i++)
				{
					if (val == binHeader[i].Max)
					{
						val = i+1;
						break;
					}
				}
			}
			catch
			{
				System.Diagnostics.Debug.Assert(false,"Exception thrown in ConfusionMatrixView.FindHeaderIdx");
			}
			return val;
		}

		/// <summary>
		/// Return -1 if a row/column total, or top/left (home) cell
		/// </summary>
		/// <param name="binHeader">Row or Column Header array</param>
		/// <param name="idx">Index to that array</param>
		/// <returns>Row / Column header number (class name, etc...)</returns>
		int FindHeaderFromIdx( SummarizeBin[] binHeader, int idx )
		{
			// return ( (idx == binHeader.Length) || (idx == -1) ) ? -1 : (int)(binHeader[idx].Max);
			int len = binHeader.Length;
			if ( (len == idx) || (idx == -1) )
				return -1;
			else
				return (int)(binHeader[idx].Max);
		}
		
		#endregion
	}

	public class ConfusionMatrixViewPropertyObj
	{
		public ConfusionMatrixViewPropertyObj ( )
		{
		}
		private string contextKey = "";
		public string ContextKey
		{
			get
			{
				return (this.contextKey);
			}
			set
			{				
				this.contextKey = value;								
			}
		}
		// Use this property to distinguish between this item in the Main Review view,
		// the PWQ View, the Advanced Classify View, DBB View, etc...
		private string viewParent = "";
		public string ViewParent
		{
			get { return (this.viewParent); }
			set { this.viewParent = value; }
		}
	}
}
